#!/usr/bin/env python
#
#   Copyright 2015 Dan Smith (dsmith+uvc@danplanet.com)
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.


import json
import os
import pprint
import urllib.parse as urlparse
import zlib
from http import client as httplib
from typing import Any, Literal

from uvcclient.const import LOGGER


class Invalid(Exception):
    pass


class NotAuthorized(Exception):
    pass


class NvrError(Exception):
    pass


class CameraConnectionError(Exception):
    pass


class UVCRemote:
    """Remote control client for Ubiquiti Unifi Video NVR."""

    CHANNEL_NAMES = ["high", "medium", "low"]

    def __init__(
        self, host: str, port: int, apikey: str, path: str = "/", ssl: bool = False
    ) -> None:
        self._host = host
        self._port = port
        self._path = path
        self._ssl = ssl
        if path != "/":
            raise Invalid("Path not supported yet")
        self._apikey = apikey
        self._bootstrap = self._get_bootstrap()
        version = ".".join(str(x) for x in self.server_version)
        LOGGER.debug(f"Server version is {version}")

    @property
    def server_version(self) -> tuple[int, int, int]:
        version = self._bootstrap["systemInfo"]["version"].split(".")
        major = int(version[0])
        minor = int(version[1])
        try:
            rev = int(version[2])
        except ValueError:
            rev = 0
        return (major, minor, rev)

    @property
    def camera_identifier(self) -> str:
        if self.server_version >= (3, 2, 0):
            return "id"
        else:
            return "uuid"

    def _get_http_connection(self) -> httplib.HTTPConnection:
        if self._ssl:
            return httplib.HTTPSConnection(self._host, self._port)
        else:
            return httplib.HTTPConnection(self._host, self._port)

    def _safe_request(self, *args: Any, **kwargs: Any) -> httplib.HTTPResponse:
        try:
            conn = self._get_http_connection()
            conn.request(*args, **kwargs)
            return conn.getresponse()
        except OSError as ex:
            raise CameraConnectionError("Unable to contact camera") from ex
        except httplib.HTTPException as ex:
            raise CameraConnectionError(f"Error connecting to camera: {ex!s}") from ex

    def _uvc_request(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        try:
            return self._uvc_request_safe(*args, **kwargs)
        except OSError as ex:
            raise NvrError("Failed to contact NVR") from ex
        except httplib.HTTPException as ex:
            raise NvrError(f"Error connecting to camera: {ex!s}") from ex

    def _uvc_request_safe(
        self,
        path: str,
        method: str = "GET",
        data: dict[str, Any] | None = None,
        mimetype: str = "application/json",
    ) -> dict[str, Any]:
        conn = self._get_http_connection()
        if "?" in path:
            url = f"{path}&apiKey={self._apikey}"
        else:
            url = f"{path}?apiKey={self._apikey}"

        headers = {
            "Content-Type": mimetype,
            "Accept": "application/json, text/javascript, */*; q=0.01",
            "Accept-Encoding": "gzip, deflate, sdch",
        }
        LOGGER.debug(f"{method} {url} headers={headers} data={data!r}")
        body = None
        if data:
            body = json.dumps(data)
        conn.request(method, url, body, headers)
        resp = conn.getresponse()
        headers = dict(resp.getheaders())
        LOGGER.debug(f"{method} {url} Result: {resp.status} {resp.reason}")
        if resp.status in (401, 403):
            raise NotAuthorized("NVR reported authorization failure")
        if resp.status / 100 != 2:
            raise NvrError(f"Request failed: {resp.status}")

        res = resp.read()
        if (
            headers.get("content-encoding") == "gzip"
            or headers.get("Content-Encoding") == "gzip"
        ):
            res = zlib.decompress(res, 32 + zlib.MAX_WBITS)
        return json.loads(res.decode())

    def _get_bootstrap(self) -> dict[str, Any]:
        return self._uvc_request("/api/2.0/bootstrap")["data"][0]

    def dump(self, uuid: str) -> None:
        """Dump information for a camera by UUID."""
        data = self._uvc_request(f"/api/2.0/camera/{uuid}")
        pprint.pprint(data)

    def set_recordmode(self, uuid: str, mode: str, chan: str | None = None) -> bool:
        """
        Set the recording mode for a camera by UUID.

        :param uuid: Camera UUID
        :param mode: One of none, full, or motion
        :param chan: One of the values from CHANNEL_NAMES
        :returns: True if successful, False or None otherwise
        """
        url = f"/api/2.0/camera/{uuid}"
        data = self._uvc_request(url)
        settings = data["data"][0]["recordingSettings"]
        mode = mode.lower()
        if mode == "none":
            settings["fullTimeRecordEnabled"] = False
            settings["motionRecordEnabled"] = False
        elif mode == "full":
            settings["fullTimeRecordEnabled"] = True
            settings["motionRecordEnabled"] = False
        elif mode == "motion":
            settings["fullTimeRecordEnabled"] = False
            settings["motionRecordEnabled"] = True
        else:
            raise Invalid("Unknown mode")

        if chan:
            settings["channel"] = self.CHANNEL_NAMES.index(chan)

        data = self._uvc_request(url, "PUT", json.dumps(data["data"][0]))
        updated = data["data"][0]["recordingSettings"]
        return settings == updated

    def get_recordmode(self, uuid: str) -> Literal["none", "full", "motion"]:
        url = f"/api/2.0/camera/{uuid}"
        data = self._uvc_request(url)
        recmodes = data["data"][0]["recordingSettings"]
        if recmodes["fullTimeRecordEnabled"]:
            return "full"
        elif recmodes["motionRecordEnabled"]:
            return "motion"
        else:
            return "none"

    def get_picture_settings(self, uuid: str) -> dict[str, Any]:
        url = f"/api/2.0/camera/{uuid}"
        data = self._uvc_request(url)
        return data["data"][0]["ispSettings"]

    def set_picture_settings(
        self, uuid: str, settings: dict[str, Any]
    ) -> dict[str, Any]:
        url = f"/api/2.0/camera/{uuid}"
        data = self._uvc_request(url)
        for key in settings:
            dtype = type(data["data"][0]["ispSettings"][key])
            try:
                data["data"][0]["ispSettings"][key] = dtype(settings[key])
            except ValueError as ex:
                raise Invalid(
                    f"Setting `{key}' requires {dtype.__name__} not {type(settings[key]).__name__}"
                ) from ex
        data = self._uvc_request(url, "PUT", json.dumps(data["data"][0]))
        return data["data"][0]["ispSettings"]

    def prune_zones(self, uuid: str) -> None:
        url = f"/api/2.0/camera/{uuid}"
        data = self._uvc_request(url)
        data["data"][0]["zones"] = [data["data"][0]["zones"][0]]
        self._uvc_request(url, "PUT", json.dumps(data["data"][0]))

    def list_zones(self, uuid: str) -> list[dict[str, Any]]:
        url = f"/api/2.0/camera/{uuid}"
        data = self._uvc_request(url)
        return data["data"][0]["zones"]

    def index(self) -> list[dict[str, Any]]:
        """
        Return an index of available cameras.

        :returns: A list of dictionaries with keys of name, uuid
        """
        cams = self._uvc_request("/api/2.0/camera")["data"]
        return [
            {
                "name": x["name"],
                "uuid": x["uuid"],
                "state": x["state"],
                "managed": x["managed"],
                "id": x["_id"],
            }
            for x in cams
            if not x["deleted"]
        ]

    def name_to_uuid(self, name: str) -> str | None:
        """
        Attempt to convert a camera name to its UUID.

        :param name: Camera name
        :returns: The UUID of the first camera with the same name if found,
                  otherwise None. On v3.2.0 and later, returns id.
        """
        cameras = self.index()
        if self.server_version >= (3, 2, 0):
            cams_by_name = {x["name"]: x["id"] for x in cameras}
        else:
            cams_by_name = {x["name"]: x["uuid"] for x in cameras}
        return cams_by_name.get(name)

    def get_camera(self, uuid: str) -> dict[str, Any]:
        return self._uvc_request(f"/api/2.0/camera/{uuid}")["data"][0]

    def get_snapshot(self, uuid: str) -> bytes:
        url = f"/api/2.0/snapshot/camera/{uuid}?force=true&apiKey={self._apikey}"
        print(url)
        resp = self._safe_request("GET", url)
        if resp.status != 200:
            raise NvrError("Snapshot returned %i" % resp.status)
        return resp.read()


def get_auth_from_env() -> tuple[str | None, int, str | None, str]:
    """
    Attempt to get UVC NVR connection information from the environment.

    Supports either a combined variable called UVC formatted like:

        UVC="http://192.168.1.1:7080/?apiKey=XXXXXXXX"

    or individual ones like:

        UVC_HOST=192.168.1.1
        UVC_PORT=7080
        UVC_APIKEY=XXXXXXXXXX

    :returns: A tuple like (host, port, apikey, path)
    """
    combined = os.getenv("UVC")
    if combined:
        # http://192.168.1.1:7080/apikey
        result = urlparse.urlparse(combined)
        if ":" in result.netloc:
            host, found_port = result.netloc.split(":", 1)
            port = int(found_port)
        else:
            host = result.netloc
            port = 7080
        apikey = urlparse.parse_qs(result.query)["apiKey"][0]
        path = result.path
        return host, port, apikey, path
    else:
        env_host = os.getenv("UVC_HOST")
        env_port = int(os.getenv("UVC_PORT", 7080))
        env_apikey = os.getenv("UVC_APIKEY")
        env_path = "/"
        return env_host, env_port, env_apikey, env_path



#
#   Copyright 2015 Dan Smith (dsmith+uvc@danplanet.com)
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import getpass
import logging
import optparse
import sys
from typing import Any

from . import camera, nvr, store
from .nvr import Invalid, UVCRemote

INFO_STORE = store.get_info_store()


def do_led(camera_info: dict[str, Any], enabled: bool) -> None:
    password = INFO_STORE.get_camera_password(camera_info["uuid"]) or "ubnt"
    cam_client = camera.UVCCameraClient(
        camera_info["host"], camera_info["username"], password
    )
    cam_client.login()
    cam_client.set_led(enabled)


def do_snapshot(client: UVCRemote, camera_info: dict[str, Any]) -> bytes:
    password = INFO_STORE.get_camera_password(camera_info["uuid"]) or "ubnt"
    cam_client: camera.UVCCameraClient
    if client.server_version >= (3, 2, 0):
        cam_client = camera.UVCCameraClientV320(
            camera_info["host"], camera_info["username"], password
        )
    else:
        cam_client = camera.UVCCameraClient(
            camera_info["host"], camera_info["username"], password
        )
    try:
        cam_client.login()
        return cam_client.get_snapshot()
    except (camera.CameraAuthError, camera.CameraConnectError):
        # Fall back to proxy through the NVR
        return client.get_snapshot(camera_info["uuid"])


def do_reboot(client: UVCRemote, camera_info: dict[str, Any]) -> None:
    password = INFO_STORE.get_camera_password(camera_info["uuid"]) or "ubnt"
    cam_client: camera.UVCCameraClient
    if client.server_version >= (3, 2, 0):
        cam_client = camera.UVCCameraClientV320(
            camera_info["host"], camera_info["username"], password
        )
    else:
        cam_client = camera.UVCCameraClient(
            camera_info["host"], camera_info["username"], password
        )
    try:
        cam_client.login()
        return cam_client.reboot()
    except camera.CameraAuthError:
        print("Failed to login to camera")
    except camera.CameraConnectError:
        print("Failed to connect to camera")
    except Exception as e:
        print(f"Failed to reboot: {e}")


def do_set_password(opts: optparse.Values) -> None:
    print("This will store the administrator password for a camera ")
    print("for later use. It will be stored on disk obscured, but ")
    print("NOT ENCRYPTED! If this is not okay, cancel now.")
    print("")
    password1 = getpass.getpass("Password: ")
    password2 = getpass.getpass("Confirm: ")
    if password1 != password2:
        print("Passwords do not match")
        return
    INFO_STORE.set_camera_password(opts.uuid, password1)
    print("Password set")


def main() -> int:
    host, port, apikey, path = nvr.get_auth_from_env()

    parser = optparse.OptionParser()
    parser.add_option("-H", "--host", default=host, help="UVC Hostname")
    parser.add_option("-P", "--port", default=port, type=int, help="UVC Port")
    parser.add_option("-K", "--apikey", default=apikey, help="UVC API Key")
    parser.add_option("-v", "--verbose", action="store_true", default=False)
    parser.add_option("-d", "--dump", action="store_true", default=False)
    parser.add_option("-u", "--uuid", default=None, help="Camera UUID")
    parser.add_option("--name", default=None, help="Camera name")
    parser.add_option("-l", "--list", action="store_true", default=False)
    parser.add_option(
        "--recordmode", default=None, help="Recording mode (none,full,motion)"
    )
    parser.add_option(
        "--get-recordmode",
        default=None,
        action="store_true",
        help="Show recording mode",
    )
    parser.add_option(
        "--recordchannel", default=None, help="Recording channel (high,medium,low)"
    )
    parser.add_option(
        "-p",
        "--get-picture-settings",
        action="store_true",
        default=False,
        help="Return picture settings as a string",
    )
    parser.add_option(
        "--set-picture-settings",
        default=None,
        help=(
            "Set picture settings with a string like that "
            "returned from --get-picture-settings"
        ),
    )
    parser.add_option(
        "--set-led",
        default=None,
        metavar="ENABLED",
        help="Enable/Disable front LED (on,off)",
    )
    parser.add_option(
        "--get-snapshot",
        default=None,
        action="store_true",
        help="Get a snapshot image and write to stdout",
    )
    parser.add_option(
        "--reboot", default=None, action="store_true", help="Reboot camera"
    )
    parser.add_option(
        "--prune-zones",
        default=None,
        action="store_true",
        help="Prune all but the first motion zone",
    )
    parser.add_option(
        "--list-zones", default=None, action="store_true", help="List motion zones"
    )
    parser.add_option(
        "--set-password",
        default=None,
        action="store_true",
        help="Store camera password",
    )
    opts, args = parser.parse_args()

    if not all([host, port, apikey]):
        print("Host, port, and apikey are required")
        return 1

    if opts.verbose:
        level = logging.DEBUG
    else:
        level = logging.WARNING
    logging.basicConfig(level=level)

    client = nvr.UVCRemote(opts.host, opts.port, opts.apikey)

    if opts.name:
        opts.uuid = client.name_to_uuid(opts.name)
        if not opts.uuid:
            print(f"`{opts.name}' is not a valid name")
            return 1

    if opts.dump:
        client.dump(opts.uuid)
    elif opts.list:
        for cam in client.index():
            ident = cam[client.camera_identifier]
            recmode = client.get_recordmode(ident)
            if not cam["managed"]:
                status = "new"
            elif cam["state"] == "FIRMWARE_OUTDATED":
                status = "outdated"
            elif cam["state"] == "UPGRADING":
                status = "upgrading"
            elif cam["state"] == "DISCONNECTED":
                status = "offline"
            elif cam["state"] == "CONNECTED":
                status = "online"
            else:
                status = "unknown:{}".format(cam["state"])
            print(
                "%s: %-24.24s [%10s] %s" % (cam["uuid"], cam["name"], status, recmode)
            )
    elif opts.recordmode:
        if not opts.uuid:
            print("Name or UUID is required")
            return 1

        r = client.set_recordmode(opts.uuid, opts.recordmode, opts.recordchannel)
        if r is True:
            return 0
        else:
            return 1
    elif opts.get_recordmode:
        if not opts.uuid:
            print("Name or UUID is required")
            return 1
        res = client.get_recordmode(opts.uuid)
        print(res)
        return res == "none"
    elif opts.get_picture_settings:
        settings = client.get_picture_settings(opts.uuid)
        print(",".join([f"{k}={v}" for k, v in settings.items()]))
        return 0
    elif opts.set_picture_settings:
        settings = {}
        try:
            for setting in opts.set_picture_settings.split(","):
                k, v = setting.split("=")
                settings[k] = v
        except ValueError:
            print("Invalid picture setting string format")
            return 1
        try:
            result = client.set_picture_settings(opts.uuid, settings)
        except Invalid as e:
            print(f"Invalid value: {e}")
            return 1
        for k in settings:
            if type(result[k])(settings[k]) != result[k]:
                print(f"Rejected: {k}")
        return 0
    elif opts.set_led is not None:
        camera = client.get_camera(opts.uuid)
        if not camera:
            print("No such camera")
            return 1
        if "Micro" not in camera["model"]:
            print("Only micro cameras support LED status")
            return 2
        do_led(camera, opts.set_led.lower() == "on")
    elif opts.prune_zones:
        if not opts.uuid:
            print("Name or UUID is required")
            return 1
        client.prune_zones(opts.uuid)
    elif opts.list_zones:
        if not opts.uuid:
            print("Name or UUID is required")
            return 1
        zones = client.list_zones(opts.uuid)
        for zone in zones:
            print(zone["name"])
    elif opts.get_snapshot:
        camera = client.get_camera(opts.uuid)
        if not camera:
            print("No such camera")
            return 1
        sys.stdout.buffer.write(do_snapshot(client, camera))
    elif opts.reboot:
        camera = client.get_camera(opts.uuid)
        if not camera:
            print("No such camera")
            return 1
        do_reboot(client, camera)
    elif opts.set_password:
        do_set_password(opts)
    else:
        print("No action specified; try --help")
    return 0



#   Copyright 2015 Dan Smith (dsmith+uvc@danplanet.com)
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import urllib.parse as urlparse
from http import client as httplib
from typing import Any

from uvcclient.const import LOGGER


class CameraConnectError(Exception):
    pass


class CameraAuthError(Exception):
    pass


class UVCCameraClient:
    def __init__(self, host: str, username: str, password: str, port: int = 80) -> None:
        self._host = host
        self._port = port
        self._username = username
        self._password = password
        self._cookie = ""

    def _safe_request(self, *args: Any, **kwargs: Any) -> httplib.HTTPResponse:
        try:
            conn = httplib.HTTPConnection(self._host, self._port)
            conn.request(*args, **kwargs)
            return conn.getresponse()
        except OSError as ex:
            raise CameraConnectError("Unable to contact camera") from ex
        except httplib.HTTPException as ex:
            raise CameraConnectError(f"Error connecting to camera: {ex!s}") from ex

    def login(self) -> None:
        resp = self._safe_request("GET", "/")
        headers = dict(resp.getheaders())
        try:
            self._cookie = headers["Set-Cookie"]
        except KeyError:
            self._cookie = headers["set-cookie"]
        session = self._cookie.split("=")[1].split(";")[0]

        urlencode = urlparse.urlencode

        data = urlencode(
            {
                "username": self._username,
                "password": self._password,
                "AIROS_SESSIONID": session,
            }
        )
        headers = {
            "Content-type": "application/x-www-form-urlencoded",
            "Accept": "*",
            "Cookie": self._cookie,
        }
        resp = self._safe_request("POST", "/login.cgi", data, headers)
        if resp.status != 200:
            raise CameraAuthError(f"Failed to login: {resp.reason}")

    def _cfgwrite(self, setting: str, value: str | int) -> bool:
        headers = {"Cookie": self._cookie}
        resp = self._safe_request(
            "GET", f"/cfgwrite.cgi?{setting}={value}", headers=headers
        )
        LOGGER.debug(f"Setting {setting}={value}: {resp.status} {resp.reason}")
        return resp.status == 200

    def set_led(self, enabled: bool) -> bool:
        return self._cfgwrite("led.front.status", int(enabled))

    @property
    def snapshot_url(self) -> str:
        return "/snapshot.cgi"

    @property
    def reboot_url(self) -> str:
        return "/api/1.1/reboot"

    @property
    def status_url(self) -> str:
        return "/api/1.1/status"

    def get_snapshot(self) -> bytes:
        headers = {"Cookie": self._cookie}
        resp = self._safe_request("GET", self.snapshot_url, headers=headers)
        if resp.status in (401, 403, 302):
            raise CameraAuthError("Not logged in")
        elif resp.status != 200:
            raise CameraConnectError(f"Snapshot failed: {resp.status}")
        return resp.read()

    def reboot(self) -> None:
        headers = {"Cookie": self._cookie}
        resp = self._safe_request("GET", self.reboot_url, headers=headers)
        if resp.status in (401, 403, 302):
            raise CameraAuthError("Not logged in")
        elif resp.status != 200:
            raise CameraConnectError(f"Reboot failed: {resp.status}")

    def get_status(self) -> dict[str, Any]:
        headers = {"Cookie": self._cookie}
        resp = self._safe_request("GET", self.status_url, headers=headers)
        if resp.status in (401, 403, 302):
            raise CameraAuthError("Not logged in")
        elif resp.status != 200:
            raise CameraConnectError(f"Status failed: {resp.status}")
        return json.loads(resp.read().decode())


class UVCCameraClientV320(UVCCameraClient):
    @property
    def snapshot_url(self) -> str:
        return "/snap.jpeg"

    def login(self) -> None:
        headers = {"Content-Type": "application/json"}
        data = json.dumps({"username": self._username, "password": self._password})
        resp = self._safe_request("POST", "/api/1.1/login", data, headers=headers)
        if resp.status != 200:
            raise CameraAuthError(f"Failed to login: {resp.reason}")
        headers = dict(resp.getheaders())
        try:
            self._cookie = headers["Set-Cookie"]
        except KeyError:
            self._cookie = headers["set-cookie"]



import base64
import json
import logging
import os
from typing import Any

LOG = logging.getLogger(__name__)
_INFO_STORE = None


class UnableToManageStore(Exception):
    pass


class InfoStore:
    _data: dict[str, Any]

    def __init__(self, path: str | None = None) -> None:
        if path is None:
            path = os.path.expanduser(os.path.join("~", ".uvcclient"))
        self._path = path
        self.load()

    def load(self) -> None:
        try:
            with open(self._path) as f:
                self._data = json.loads(base64.b64decode(f.read()).decode())
        except OSError:
            LOG.debug("No info store")
            self._data = {}
        except Exception as ex:
            LOG.error("Failed to read store data: %s", ex)
            raise UnableToManageStore("Unable to write to store") from ex

    def save(self) -> None:
        try:
            with open(self._path, "w") as f:
                f.write(base64.b64encode(json.dumps(self._data).encode()).decode())
            os.chmod(self._path, 0o600)
        except OSError as ex:
            LOG.error("Unable to write store: %s", str(ex))
            raise UnableToManageStore("Unable to write to store") from ex

    def get_camera_passwords(self) -> dict[str, str]:
        return self._data.get("camera_passwords", {})

    def get_camera_password(self, uuid: str) -> str | None:
        return self.get_camera_passwords().get(uuid)

    def set_camera_password(self, uuid: str, password: str) -> None:
        if "camera_passwords" not in self._data:
            self._data["camera_passwords"] = {}
        self._data["camera_passwords"][uuid] = password
        self.save()


def get_info_store(path: str | None = None) -> InfoStore:
    global _INFO_STORE
    if _INFO_STORE is None:
        _INFO_STORE = InfoStore(path)
    return _INFO_STORE
